/*CLAX.C**********************************************************************

   Clax: Portable keyframing library
         see inclosed LICENSE.TXT for licensing terms.

         for documentation, refer to CLAX.TXT

   author           : Borzom
   file created     : 17/04/97
   file description : main clax routines

   revision history :
     v0.10 (17/04/97) Borzom: Initial version.

   notes            :
     the library itself, if it has bugs, its probably here :)

*****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include "clax.h"
#include "claxi.h"

/*****************************************************************************
  copyright, format table, etc...
*****************************************************************************/

  char    clax_version[]   = "clax version 0.10(beta)";
  char    clax_copyright[] = "copyright (c) 1997 borzom";

  c_SCENE *clax_scene; /* current active screne */
  int      clax_flags; /* curreng flags         */

struct {
  char *name;                    /* file extension        */
  int  (*load_mesh)   (FILE *f); /* loads mesh            */
  int  (*load_motion) (FILE *f); /* loads motion          */
  int  (*save_scene)  (FILE *f); /* saves the whole scene */
} clax_drivers[] = {
  {"3DS", clax_load_mesh_3DS, clax_load_motion_3DS, NULL},
//{"ASC", clax_load_mesh_ASC, NULL,                 NULL},
//{"DXF", clax_load_mesh_DXF, NULL,                 NULL},
//{"CLX", clax_load_mesh_CLX, clax_load_motion_CLX, clax_save_CLX}
};

/*****************************************************************************
  some internal stuff
*****************************************************************************/

static void clax_free_key (t_KEY *key)
{
/*
  clax_free_key: frees keys list.
*/
  t_KEY *next;

  for (; key; key = next) {
    next = key->next;
    free (key);
  }
}

static void calc_normals ()
{
/*
  calc_normals: calculates face/vertex normals.
*/
  c_OBJECT *obj;
  w_NODE   *node;
  c_VECTOR  a, b, normal;
  int       i, j, num;

  for (node = clax_scene->world; node; node = node->next) {
    if (node->type == clax_obj_object) {
      obj = (c_OBJECT *)node->object;
      for (i = 0; i < obj->numfaces; i++) { /* face normals */
        vec_sub (&obj->vert_orig [obj->faces[i].a].vert,
                 &obj->vert_orig [obj->faces[i].b].vert, &a);
        vec_sub (&obj->vert_orig [obj->faces[i].b].vert,
                 &obj->vert_orig [obj->faces[i].c].vert, &b);
        vec_cross (&a, &b, &normal);
        vec_normalize (&normal, &obj->faces[i].norm);
      }
      for (i = 0; i < obj->numverts; i++) { /* vertex normals */
        num = 0;
        vec_zero (&normal);
        for (j = 0; j < obj->numfaces; j++) {
          if (obj->faces[j].a == i ||
              obj->faces[j].b == i ||
              obj->faces[j].c == i) {
            vec_add (&normal, &obj->faces[j].norm, &normal);
            num++;
          }
        }
        if (num) vec_scale (&normal, 1.0 / (float)num, &normal);
        vec_normalize (&normal, &obj->vert_orig[i].norm);
      }
    }
  }
}

static void do_transform ()
{
/*
  do_transform: transform all objects.
*/
  w_NODE   *node;
  c_OBJECT *obj;
  c_VECTOR  vec;
  int       i, j;

  for (node = clax_scene->world; node; node = node->next) {
    if (node->type == clax_obj_object) {
      obj = (c_OBJECT *)node->object;
      for (i = 0; i < obj->numverts; i++) {
        vec_sub (&obj->vert_orig[i].vert, &obj->pivot, &vec);
        mat_mulvec (obj->matrix, &vec, &obj->vert_trans[i].vert);
        mat_mulvec (obj->matrix, &obj->vert_trans[i].norm,
                                 &obj->vert_trans[i].norm);
      }
    }
  }
}

static int strucmp (char *s1, char *s2)
{
/*
  strucmp: non-case sensitive string compare.
*/
  int diff = 0;

  do {
    diff += (toupper (*s1) - toupper (*s2));
  } while (*s1++ && *s2++);
  return diff;
}

/*****************************************************************************
  the library itself.
*****************************************************************************/

int clax_init (int flags)
{
/*
  clax_init: clax initialization.
*/
  clax_scene = NULL;
  clax_flags = flags;
  return clax_err_ok;
}

int clax_done ()
{
/*
  clax_done: clax deinitialization.
*/
  clax_scene = NULL;
  clax_flags = 0;
  return clax_err_ok;
}

char *clax_geterror (int code)
{
/*
  clax_geterror: return error string.
*/
  switch (code) {
    case clax_err_nomem:     return "not enough memory";
    case clax_err_nofile:    return "file not found";
    case clax_err_badfile:   return "corrupted file";
    case clax_err_badver:    return "unsupported version";
    case clax_err_badformat: return "unsupported format";
    case clax_err_badframe:  return "invalid frame number";
    case clax_err_badname:   return "invalid object name";
    case clax_err_noframes:  return "no frames in keyframer";
    case clax_err_notloaded: return "scene not loaded";
    case clax_err_nullptr:   return "null pointer assignment";
    case clax_err_spline:    return "less than 2 keys in spline";
    case clax_err_singular:  return "cannot inverse singlular matrix";
    case clax_err_badid:     return "bad object id";
    case clax_err_exist:     return "object already exist";
    case clax_err_undefined:
    default:                 return "internal error";
  }
}

int clax_getframes (float *start, float *end)
{
/*
  clax_getframes: return number of frames.
*/
  if (!clax_scene) return clax_err_notloaded;
  if (!clax_scene->keyframer) return clax_err_notloaded;
  if (clax_scene->f_end - clax_scene->f_start == 0) return clax_err_noframes;
  *start = clax_scene->f_start;
  *end = clax_scene->f_end;
  return clax_err_ok;
}

int clax_setframe (float frame)
{
/*
  clax_setframe: set current frame number.
*/
  if (!clax_scene) return clax_err_notloaded;
  if (!clax_scene->keyframer) return clax_err_notloaded;
  if (clax_scene->f_end - clax_scene->f_start == 0) return clax_err_noframes;
  if (clax_scene->f_start <= frame && clax_scene->f_end >= frame) {
    clax_scene->f_current = frame;
    return clax_err_ok;
  } else return clax_err_badframe;
}

int clax_getframe (float *frame)
{
/*
  clax_getframe: return current frame number.
*/
  if (!clax_scene) return clax_err_notloaded;
  if (!clax_scene->keyframer) return clax_err_notloaded;
  if (clax_scene->f_end - clax_scene->f_start == 0) return clax_err_noframes;
  *frame = clax_scene->f_current;
  return clax_err_ok;
}

int clax_setactive (c_SCENE *scene)
{
/*
  clax_setactivescene: sets active scene.
*/
  if (scene == NULL) return clax_err_nullptr;
  clax_scene = scene;
  return clax_err_ok;
}

int clax_getactive (c_SCENE **scene)
{
/*
  clax_getactivescene: get active scene.
*/
  *scene = clax_scene;
  return clax_err_ok;
}

int clax_byname (char *name, w_NODE **node)
{
/*
  clax_byname: find object by name.
*/
  if (!clax_scene) return clax_err_notloaded;
  if (!clax_scene->world) return clax_err_notloaded;
  for (*node = clax_scene->world; *node; *node = (*node)->next) {
    switch ((*node)->type) {
      case clax_obj_camera:
        if (strcmp (((c_CAMERA *)((*node)->object))->name, name) == 0)
          return clax_err_ok; break;
      case clax_obj_object:
        if (strcmp (((c_OBJECT *)((*node)->object))->name, name) == 0)
          return clax_err_ok; break;
      case clax_obj_light:
        if (strcmp (((c_LIGHT *)((*node)->object))->name, name) == 0)
          return clax_err_ok; break;
      case clax_obj_material:
        if (strcmp (((c_MATERIAL *)((*node)->object))->name, name) == 0)
          return clax_err_ok; break;
      case clax_obj_ambient:
        if (strcmp (((c_AMBIENT *)((*node)->object))->name, name) == 0)
          return clax_err_ok;
    }
  }
  return clax_err_ok; /* return NULL */
}

int clax_byid (int id, w_NODE **node)
{
/*
  clax_byid: find object by id.
*/
  if (!clax_scene) return clax_err_notloaded;
  if (!clax_scene->world) return clax_err_notloaded;
  for (*node = clax_scene->world; *node; *node = (*node)->next) {
    switch ((*node)->type) {
      case clax_obj_camera:
        if (((c_CAMERA *)((*node)->object))->id == id) return clax_err_ok;
        break;
      case clax_obj_object:
        if (((c_OBJECT *)((*node)->object))->id == id) return clax_err_ok;
        break;
      case clax_obj_light:
        if (((c_LIGHT *)((*node)->object))->id == id) return clax_err_ok;
        break;
      case clax_obj_material:
        if (((c_MATERIAL *)((*node)->object))->id == id) return clax_err_ok;
        break;
      case clax_obj_ambient:
        if (((c_AMBIENT *)((*node)->object))->id == id) return clax_err_ok;
    }
  }
  return clax_err_ok; /* return NULL */
}

int clax_findfirst (int attr, w_NODE **node)
{
/*
  clax_findfirst: finds first node with attribute "attr".
*/
  if (!clax_scene) return clax_err_notloaded;
  if (!clax_scene->world) return clax_err_notloaded;
  for (*node = clax_scene->world; *node; *node = (*node)->next)
    if ((*node)->type & attr) return clax_err_ok;
  return clax_err_ok; /* return NULL */
}

int clax_findnext (int attr, w_NODE **node)
{
/*
  clax_findnext : finds next node with attribute "attr".
*/
  if (!clax_scene) return clax_err_notloaded;
  if (!clax_scene->world) return clax_err_notloaded;
  for (*node = (*node)->next; *node; *node = (*node)->next)
    if ((*node)->type & attr) return clax_err_ok;
  return clax_err_ok; /* return NULL */
}

int clax_add_world (int type, void *obj)
{
/*
  clax_add_world: add object to world list.
*/
  w_NODE *node;

  if (!clax_scene) return clax_err_notloaded;
  if ((node = (w_NODE *)malloc (sizeof (w_NODE))) == NULL)
    return clax_err_nomem;
  node->type = type;
  node->object = obj;
  node->next = NULL;
  if (clax_scene->world == NULL) {
    node->prev = NULL;
    clax_scene->world = node;
    clax_scene->wtail = node;
  } else {
    node->prev = clax_scene->wtail;
    clax_scene->wtail->next = node;
    clax_scene->wtail = node;
  }
  return clax_err_ok;
}

int clax_add_track (int type, int id, int parent, void *track, void *obj)
{
/*
  clax_add_track: add track to keyframer list.
*/
  k_NODE *node, *pnode;

  if (!clax_scene) return clax_err_notloaded;
  if ((node = (k_NODE *)malloc (sizeof (k_NODE))) == NULL)
    return clax_err_nomem;
  node->type = type;
  node->id = id;
  node->track = track;
  node->brother = NULL;
  node->child = NULL;
  node->next = NULL;
  node->object = obj;
  if (clax_scene->keyframer == NULL) {
    node->prev = NULL;
    clax_scene->keyframer = node;
    clax_scene->ktail = node;
  } else {
    node->prev = clax_scene->ktail;
    clax_scene->ktail->next = node;
    clax_scene->ktail = node;
  }
  if (parent != -1) {
    for (pnode = clax_scene->keyframer; pnode; pnode = pnode->next)
      if (pnode->id == parent) {
        if (pnode->child == NULL) pnode->child = node; else {
          node->brother = pnode->child;
          pnode->child = node;
        }
      }
  }
  return clax_err_ok;
}

int clax_set_track (int type, int id, t_KEY *track)
{
/*
  clax_set_track: assigns a key list to track.
*/
  int     err;
  void   *obj;
  k_NODE *node = clax_scene->keyframer;

  if (!clax_scene) return clax_err_notloaded;
  while (node && node->id != id) node = node->next;
  if (!node) return clax_err_undefined;
  obj = node->track;

  switch (node->type) {
    case clax_track_camera:
      switch (type) {
        case clax_key_pos:  ((t_CAMERA *)obj)->pos = track; break;
        case clax_key_fov:  ((t_CAMERA *)obj)->fov = track; break;
        case clax_key_roll: ((t_CAMERA *)obj)->roll = track;
      }
      break;
    case clax_track_cameratgt:
      switch (type) {
        case clax_key_pos: ((t_CAMERATGT *)obj)->pos = track;
      }
      break;
    case clax_track_light:
      switch (type) {
        case clax_key_pos:   ((t_LIGHT *)obj)->pos = track; break;
        case clax_key_color: ((t_LIGHT *)obj)->color = track;
      }
      break;
    case clax_track_lighttgt:
      switch (type) {
        case clax_key_pos:   ((t_LIGHTTGT *)obj)->pos = track; break;
      }
    case clax_track_spotlight:
      switch (type) {
        case clax_key_pos:   ((t_SPOTLIGHT *)obj)->pos = track; break;
        case clax_key_color: ((t_SPOTLIGHT *)obj)->color = track; break;
        case clax_key_roll:  ((t_SPOTLIGHT *)obj)->roll = track;
      }
      break;
    case clax_track_object:
      switch (type) {
        case clax_key_pos:    ((t_OBJECT *)obj)->translate = track; break;
        case clax_key_scale:  ((t_OBJECT *)obj)->scale = track; break;
        case clax_key_rotate: ((t_OBJECT *)obj)->rotate = track; break;
        case clax_key_morph:  ((t_OBJECT *)obj)->morph = track; break;
        case clax_key_hide:   ((t_OBJECT *)obj)->hide = track;
      }
      break;
    case clax_track_ambient:
      switch (type) {
        case clax_key_color: ((t_AMBIENT *)obj)->color = track;
      }
  }
  return clax_err_ok;
}

int clax_allocscene (c_SCENE **scene)
{
/*
  clax_allocscene: allocates a new scene.
*/
  if ((*scene = (c_SCENE *)malloc (sizeof (c_SCENE))) == NULL)
    return clax_err_nomem;
  (*scene)->world = NULL;
  (*scene)->wtail = NULL;
  (*scene)->keyframer = NULL;
  (*scene)->ktail = NULL;
  return clax_err_ok;
}

int clax_loadworld (char *filename, c_SCENE *scene)
{
/*
  clax_loadworld: loads mesh data from 3ds file "filename"
                  into scene "scene".
*/
  c_SCENE *old_scene = clax_scene;
  int     (*loader)(FILE *f) = NULL;
  FILE    *f;
  char    *s;
  int      i, error;

  for (i = 0; i < sizeof (clax_drivers) / sizeof (clax_drivers[0]); i++) {
    s = strchr (filename, '.') + 1;
    if (strucmp (s, clax_drivers[i].name) == 0)
      loader = clax_drivers[i].load_mesh;
  }
  if (!loader) return clax_err_badformat;
  scene->f_start = 0;
  scene->f_end = 0;
  scene->f_current = 0;
  scene->world = NULL;
  scene->wtail = NULL;
  scene->keyframer = NULL;
  scene->ktail = NULL;
  if ((f = fopen (filename, "rb")) == NULL) return clax_err_nofile;
  clax_setactive (scene);
  error = loader (f);
  clax_setactive (old_scene);
  fclose (f);
  if (error) {
    clax_free_world (scene);
    return error;
  }
  if (clax_flags & clax_calcnorm) calc_normals ();
  return clax_err_ok;
}

int clax_loadmotion (char *filename, c_SCENE *scene)
{
/*
  clax_loadmotion: loads motion data from 3ds file "filename"
                   into scene "scene".
*/
  c_SCENE  *old_scene = clax_scene;
  w_NODE   *node;
  c_OBJECT *obj;
  int      (*loader) (FILE *f) = NULL;
  FILE     *f;
  char     *s;
  int       i, error;

  if (!scene->world) return clax_err_notloaded;
  for (i = 0; i < sizeof (clax_drivers) / sizeof (clax_drivers[0]); i++) {
    s = strchr (filename, '.') + 1;
    if (strucmp (s, clax_drivers[i].name) == 0)
      loader = clax_drivers[i].load_motion;
  }
  if (!loader) return clax_err_badformat;
  if ((f = fopen (filename, "rb")) == NULL) return clax_err_nofile;
  clax_setactive (scene);
  error = loader (f);
  fclose (f);
  clax_setactive (old_scene);
  if (error) {
    clax_free_motion (scene);
    return error;
  }
  if (clax_flags & clax_transform) {
    for (node = clax_scene->world; node; node = node->next) {
      if (node->type == clax_obj_object) {
        obj = (c_OBJECT *)node->object;
        if (!(obj->flags & clax_obj_dummy))
          if ((obj->vert_trans = (c_VERTEX *)malloc (sizeof (c_VERTEX)*obj->numverts)) == NULL)
            return clax_err_nomem;
      }
    }
    do_transform ();
  }
  return clax_err_ok;
}

int clax_free_world (c_SCENE *scene)
{
/*
  clax_free_world: release all memory used by world.
*/
  w_NODE     *node, *next;
  c_LIGHT    *light;
  c_CAMERA   *cam;
  c_OBJECT   *obj;
  c_MATERIAL *mat;
  c_AMBIENT  *amb;

  if (!scene) return clax_err_nullptr;
  for (node = scene->world; node; node = next) {
    switch (node->type) {
      case clax_obj_light:
        light = (c_LIGHT *)node->object;
        free (light->name);
        break;
      case clax_obj_camera:
        cam = (c_CAMERA *)node->object;
        free (cam->name);
        break;
      case clax_obj_object:
        obj = (c_OBJECT *)node->object;
        free (obj->name);
        free (obj->vert_orig);
        free (obj->vert_trans);
        free (obj->faces);
        break;
      case clax_obj_material:
        mat = (c_MATERIAL *)node->object;
        free (mat->name);
        free (mat->texture.file);
        free (mat->bump.file);
        free (mat->reflection.file);
        break;
      case clax_obj_ambient:
        amb = (c_AMBIENT *)node->object;
        free (amb->name);
    }
    next = node->next;
    free (node->object);
    free (node);
  }
  scene->world = NULL;
  scene->wtail = NULL;
  return clax_err_ok;
}

int clax_free_motion (c_SCENE *scene)
{
/*
  clax_free_motion: release all memory used by keyframer.
*/
  k_NODE      *node, *next;
  t_CAMERA    *cam;
  t_CAMERATGT *camtgt;
  t_LIGHT     *light;
  t_LIGHTTGT  *litgt;
  t_SPOTLIGHT *spotl;
  t_OBJECT    *obj;
  t_AMBIENT   *amb;

  if (!scene) return clax_err_nullptr;
  for (node = scene->keyframer; node; node = next) {
    switch (node->type) {
      case clax_track_camera:
        cam = (t_CAMERA *)node->track;
        clax_free_key (cam->pos);
        clax_free_key (cam->fov);
        clax_free_key (cam->roll);
        break;
      case clax_track_cameratgt:
        camtgt = (t_CAMERATGT *)node->track;
        clax_free_key (camtgt->pos);
        break;
      case clax_track_light:
        light = (t_LIGHT *)node->track;
        clax_free_key (light->pos);
        clax_free_key (light->color);
        break;
      case clax_track_spotlight:
        spotl = (t_SPOTLIGHT *)node->track;
        clax_free_key (spotl->pos);
        clax_free_key (spotl->color);
        clax_free_key (spotl->roll);
        break;
      case clax_track_lighttgt:
        litgt = (t_LIGHTTGT *)node->track;
        clax_free_key (litgt->pos);
        break;
      case clax_track_object:
        obj = (t_OBJECT *)node->track;
        clax_free_key (obj->translate);
        clax_free_key (obj->scale);
        clax_free_key (obj->rotate);
        clax_free_key (obj->morph);
        clax_free_key (obj->hide);
        break;
      case clax_track_ambient:
        amb = (t_AMBIENT *)node->track;
        clax_free_key (amb->color);
    }
    next = node->next;
    free (node->track);
    free (node);
  }
  scene->keyframer = NULL;
  scene->ktail = NULL;
  return clax_err_ok;
}

int clax_loadscene (char *filename, c_SCENE *scene)
{
/*
  clax_loadscene: loads mesh and keyframer data from 3ds file "filename"
                  into scene "scene".
*/
  int error;

  if ((error = clax_loadworld (filename, scene)) != clax_err_ok)
    return error;
  return clax_loadmotion (filename, scene);
}

int clax_free_scene (c_SCENE *scene)
{
  int error;

  if ((error = clax_free_world (scene)) != clax_err_ok) return error;
  return clax_free_motion (scene);
}

int clax_getkey_float (t_KEY *keys, float frame, float *out)
{
/*
  clax_getkey_float: return float key at frame "frame".
*/
  float alpha;

  if (frame < 0.0) return clax_err_badframe;
  if (!keys) return clax_err_nullptr;
  if (!keys->next || frame < keys->frame) { /* only one key, return first key */
    *out = keys->val._float;
    return clax_err_ok;
  }
  /* more than one key, spline interpolation */
  return spline_getkey_float (keys, frame, out);
}

int clax_getkey_vect (t_KEY *keys, float frame, c_VECTOR *out)
{
/*
  clax_getkey_vect: return vector key at frame "frame".
*/
  float alpha;

  if (frame < 0.0) return clax_err_badframe;
  if (!keys) return clax_err_nullptr;
  if (!keys->next || frame < keys->frame) { /* only one key, return first key */
    vec_copy (&keys->val._vect, out);
    return clax_err_ok;
  }
  /* more than one key, spline interpolation */
  return spline_getkey_vect (keys, frame, out);
}

int clax_getkey_rgb (t_KEY *keys, float frame, c_RGB *out)
{
/*
  clax_getkey_rgb: return rgb key at frame "frame".
*/
  c_VECTOR *vect = (c_VECTOR *)out;
  return spline_getkey_vect (keys, frame, vect);
}

int clax_getkey_quat (t_KEY *keys, float frame, c_QUAT *out)
{
/*
  clax_getkey_quat: return quaternion key at frame "frame".
*/
  float alpha;

  if (frame < 0.0) return clax_err_badframe;
  if (!keys) return clax_err_nullptr;
#ifdef CLAX_QUAT_SLERP
  while (frame > keys->next->frame && keys->next) keys = keys->next;
  if (!keys->next || frame < keys->frame) { /* frame is above last key */
    qt_copy (&keys->val._quat, out);
    return clax_err_ok;
  }
  alpha = (frame - keys->frame) / (keys->next->frame - keys->frame);
  qt_slerp (&keys->val._quat, &keys->next->val._quat, alpha, out);
  return clax_err_ok;
#else
  if (!keys->next || frame < keys->frame) {
    qt_copy (&keys->val._quat, out);
    return clax_err_ok;
  }
  return spline_getkey_quat (keys, frame, out);
#endif
}

int clax_update ()
{
/*
  clax_update: update all keyframer data.
*/
  k_NODE      *node, *child;
  c_CAMERA    *cam;
  t_CAMERA    *tcam;
  t_CAMERATGT *tcamt;
  c_LIGHT     *light;
  t_LIGHT     *tlight;
  t_SPOTLIGHT *tspotl;
  t_LIGHTTGT  *tlightt;
  c_AMBIENT   *amb;
  t_AMBIENT   *tamb;
  c_OBJECT    *obj,*cobj;
  t_OBJECT    *tobj;
  c_MATRIX     c, d;
  c_QUAT       q;
  float        frame = clax_scene->f_current;

  if (!clax_scene) return clax_err_notloaded;
  /* update objects */
  for (node = clax_scene->keyframer; node; node = node->next) {
    switch (node->type) {
      case clax_track_camera:
        cam = (c_CAMERA *)node->object;
        tcam = (t_CAMERA *)node->track;
        clax_getkey_vect (tcam->pos, frame, &cam->pos);
        clax_getkey_float (tcam->fov, frame, &cam->fov);
        clax_getkey_float (tcam->roll, frame, &cam->roll);
        break;
      case clax_track_cameratgt:
        cam = (c_CAMERA *)node->object;
        tcamt = (t_CAMERATGT *)node->track;
        clax_getkey_vect (tcamt->pos, frame, &cam->target);
        break;
      case clax_track_light:
        light = (c_LIGHT *)node->object;
        tlight = (t_LIGHT *)node->track;
        clax_getkey_vect (tlight->pos, frame, &light->pos);
        clax_getkey_rgb (tlight->color, frame, &light->color);
        break;
      case clax_track_spotlight:
        light = (c_LIGHT *)node->object;
        tspotl = (t_SPOTLIGHT *)node->track;
        clax_getkey_vect (tspotl->pos, frame, &light->pos);
        clax_getkey_rgb (tspotl->color, frame, &light->color);
        clax_getkey_float (tspotl->roll, frame, &light->roll);
        break;
      case clax_track_lighttgt:
        light = (c_LIGHT *)node->object;
        tlightt = (t_LIGHTTGT *)node->track;
        clax_getkey_vect (tlightt->pos, frame, &light->target);
        break;
      case clax_track_object:
        obj = (c_OBJECT *)node->object;
        tobj = (t_OBJECT *)node->track;
        clax_getkey_vect (tobj->translate, frame, &obj->translate);
        clax_getkey_vect (tobj->scale, frame, &obj->scale);
        clax_getkey_quat (tobj->rotate, frame, &obj->rotate);
        qt_normalize (&obj->rotate, &obj->rotate);
        qt_invmatrix (&obj->rotate, c);
        mat_setscale (&obj->scale, d);
        c[X][W] = obj->translate.x;
        c[Y][W] = obj->translate.y;
        c[Z][W] = obj->translate.z;
        mat_mul (c, d, obj->matrix);
        break;
      case clax_track_ambient:
        amb = (c_AMBIENT *)node->object;
        tamb = (t_AMBIENT *)node->track;
        clax_getkey_rgb (tamb->color, frame, &amb->color);
    }
  }
  /* update hierarchy */
  if (clax_flags & clax_hierarchy) {
    for (node = clax_scene->keyframer; node; node = node->next) {
      if (node->type == clax_track_object) {
        for (child = node->child; child; child = child->brother) {
          if (child->type == clax_track_object) {
            obj = (c_OBJECT *)node->object;
            cobj = (c_OBJECT *)child->object;
            mat_mul (obj->matrix, cobj->matrix, cobj->matrix);
          }
        }
      }
    }
  }
  if (clax_flags & clax_transform) do_transform ();
  return clax_err_ok;
}
