/*MATRIX.C********************************************************************

   Clax: Portable keyframing library
         see inclosed LICENSE.TXT for licensing terms.

         for documentation, refer to CLAX.TXT

   author           : Borzom
   file created     : 17/04/97
   file description : matrix math

   revision history :
     v0.10 (17/04/97) Borzom: Initial version.

   notes            :
     ( xx , xy , xz , xw ) matrix structure.
     ( yx , yy , yz , yw )
     ( zx , zy , zz , zw )

*****************************************************************************/

#include <math.h>
#include <stdio.h>
#include "clax.h"
#include "claxi.h"

static c_MATRIX Tidentity = {
  {1, 0, 0, 0},
  {0, 1, 0, 0},
  {0, 0, 1, 0},
};

void mat_identity (c_MATRIX out)
{
/*
  mat_idenity: make identity matrix.
*/
  int i, j;

  for (i = 0; i < 3; i++)
    for (j = 0; j < 4; j++)
      out[i][j] = Tidentity[i][j];
}

void mat_zero (c_MATRIX out)
{
/*
  mat_zero: clear matrix.
*/
  int i, j;

  for (i = 0; i < 3; i++)
    for (j = 0; j < 4; j++) out[i][j] = 0.0;
}

void mat_copy (c_MATRIX a, c_MATRIX out)
{
/*
  mat_copy: matrix copy.
*/
  int i, j;

  for (i = 0; i < 3; i++)
    for (j = 0; j < 4; j++)
      out[i][j] = a[i][j];
}

void mat_print (c_MATRIX a)
{
/*
  mat_print: print matrix on stdout.
*/
  printf ("xx:%f xy:%f xz:%f xw:%f\n", a[X][X], a[X][Y], a[X][Z], a[X][W]);
  printf ("yx:%f yy:%f yz:%f yw:%f\n", a[Y][X], a[Y][Y], a[Y][Z], a[Y][W]);
  printf ("zx:%f zy:%f zz:%f zw:%f\n", a[Z][X], a[Z][Y], a[Z][Z], a[Z][W]);
}

void mat_add (c_MATRIX a, c_MATRIX b, c_MATRIX out)
{
/*
  mat_add: matrix addition.
*/
  int      i, j;
  c_MATRIX temp;
  for (i = 0; i < 3; i++)
    for (j = 0; j < 4; j++)
      temp[i][j] = a[i][j]+b[i][j];
  mat_copy (temp, out);
}

void mat_sub (c_MATRIX a, c_MATRIX b, c_MATRIX out)
{
/*
  mat_sub: matrix substraction.
*/
  int      i, j;
  c_MATRIX temp;

  for (i = 0; i < 3; i++)
    for (j = 0; j < 4; j++)
      temp[i][j] = a[i][j] - b[i][j];
  mat_copy (temp, out);
}

void mat_mul (c_MATRIX a, c_MATRIX b, c_MATRIX out)
{
/*
  mat_mul: matrix multiplication.
*/
  c_MATRIX temp;

  temp[X][X]=a[X][X]*b[X][X]+a[X][Y]*b[Y][X]+a[X][Z]*b[Z][X];
  temp[X][Y]=a[X][X]*b[X][Y]+a[X][Y]*b[Y][Y]+a[X][Z]*b[Z][Y];
  temp[X][Z]=a[X][X]*b[X][Z]+a[X][Y]*b[Y][Z]+a[X][Z]*b[Z][Z];
  temp[X][W]=a[X][X]*b[X][W]+a[X][Y]*b[Y][W]+a[X][Z]*b[Z][W]+a[X][W];
  temp[Y][X]=a[Y][X]*b[X][X]+a[Y][Y]*b[Y][X]+a[Y][Z]*b[Z][X];
  temp[Y][Y]=a[Y][X]*b[X][Y]+a[Y][Y]*b[Y][Y]+a[Y][Z]*b[Z][Y];
  temp[Y][Z]=a[Y][X]*b[X][Z]+a[Y][Y]*b[Y][Z]+a[Y][Z]*b[Z][Z];
  temp[Y][W]=a[Y][X]*b[X][W]+a[Y][Y]*b[Y][W]+a[Y][Z]*b[Z][W]+a[Y][W];
  temp[Z][X]=a[Z][X]*b[X][X]+a[Z][Y]*b[Y][X]+a[Z][Z]*b[Z][X];
  temp[Z][Y]=a[Z][X]*b[X][Y]+a[Z][Y]*b[Y][Y]+a[Z][Z]*b[Z][Y];
  temp[Z][Z]=a[Z][X]*b[X][Z]+a[Z][Y]*b[Y][Z]+a[Z][Z]*b[Z][Z];
  temp[Z][W]=a[Z][X]*b[X][W]+a[Z][Y]*b[Y][W]+a[Z][Z]*b[Z][W]+a[Z][W];
  mat_copy (temp, out);
}

int mat_inverse (c_MATRIX a, c_MATRIX out)
{
/*
  mat_inverse: inverse matrix calculation (non-singular).
*/
  float    scale;
  c_MATRIX temp;

  scale = a[X][X]*a[X][X] + a[X][Y]*a[X][Y] + a[X][Z]*a[X][Z];
  if (scale == 0.0) {
    mat_identity (out);
    return clax_err_singular;
  }
  scale = 1.0 / scale;
  temp[X][X] = scale * a[X][X];
  temp[Y][X] = scale * a[X][Y];
  temp[Z][X] = scale * a[X][Z];
  temp[X][Y] = scale * a[Y][X];
  temp[Y][Y] = scale * a[Y][Y];
  temp[Z][Y] = scale * a[Y][Z];
  temp[X][Z] = scale * a[Z][X];
  temp[Y][Z] = scale * a[Z][Y];
  temp[Z][Z] = scale * a[Z][Z];
  temp[X][W] = -(temp[X][X]*a[X][W] + temp[X][Y]*a[Y][W] + temp[X][Z]*a[Z][W]);
  temp[Y][W] = -(temp[Y][X]*a[X][W] + temp[Y][Y]*a[Y][W] + temp[Y][Z]*a[Z][W]);
  temp[Z][W] = -(temp[Z][X]*a[X][W] + temp[Z][Y]*a[Y][W] + temp[Z][Z]*a[Z][W]);
  mat_copy (temp, out);
  return clax_err_ok;
}

void mat_settrans (c_VECTOR *v, c_MATRIX out)
{
/*
  mat_settrans: create a translation matrix.
*/
  mat_identity (out);
  out[X][W] = v->x;
  out[Y][W] = v->y;
  out[Z][W] = v->z;
}

void mat_setscale (c_VECTOR *v, c_MATRIX out)
{
/*
  mat_setscale: create a scale matrix.
*/
  mat_identity (out);
  out[X][X] = v->x;
  out[Y][Y] = v->y;
  out[Z][Z] = v->z;
}

void mat_rotateX (float ang, c_MATRIX out)
{
/*
  mat_rotateX: create rotation matrix around X axis.
*/
  float sinx, cosx;

  mat_identity (out);
  sinx = sin (ang);
  cosx = cos (ang);
  out[Y][Y] =  cosx;
  out[Y][Z] =  sinx;
  out[Z][Y] = -sinx;
  out[Z][Z] =  cosx;
}

void mat_rotateY (float ang, c_MATRIX out)
{
/*
  mat_rotateY: create rotation matrix around Y axis.
*/
  float siny, cosy;

  mat_identity (out);
  siny = sin (ang);
  cosy = cos (ang);
  out[X][X] =  cosy;
  out[X][Z] = -siny;
  out[Z][X] =  siny;
  out[Z][Z] =  cosy;
}

void mat_rotateZ (float ang, c_MATRIX out)
{
/*
  mat_rotateZ: create rotation matrix around Z axis.
*/
  float sinz, cosz;

  mat_identity (out);
  sinz = sin (ang);
  cosz = cos (ang);
  out[X][X] =  cosz;
  out[X][Y] =  sinz;
  out[Y][X] = -sinz;
  out[Y][Y] =  cosz;
}

void mat_mulvec (c_MATRIX a, c_VECTOR *b, c_VECTOR *out)
{
/*
  mat_mulvec: multiply a vector by matrix (out = [a]*b)
*/
  c_VECTOR temp;

  temp.x = b->x*a[X][X] + b->y*a[X][Y] + b->z*a[X][Z] + a[X][W];
  temp.y = b->x*a[Y][X] + b->y*a[Y][Y] + b->z*a[Y][Z] + a[Y][W];
  temp.z = b->x*a[Z][X] + b->y*a[Z][Y] + b->z*a[Z][Z] + a[Z][W];
  vec_copy (&temp, out);
}
