/*QUATERN.C*******************************************************************

   Clax: Portable keyframing library
         see inclosed LICENSE.TXT for licensing terms.

         for documentation, refer to CLAX.TXT

   author           : Borzom
   file created     : 16/04/97
   file description : quaternion math

   revision history :
     v0.10 (16/04/97) Borzom: Initial version.

   notes            :

*****************************************************************************/

#include <math.h>
#include <stdio.h>
#include "clax.h"
#include "claxi.h"

void qt_fromang (float ang, float x, float y, float z, c_QUAT *out)
{
/*
  qt_fromang: compute quaterion from [angle,axis] representation.
*/
  float s, omega;

  omega = ang / 2.0;
  s = sin (omega);
  out->w = cos (omega);
  out->x = x * s;
  out->y = y * s;
  out->z = z * s;
}

void qt_toang (c_QUAT *a, float *ang, float *x, float *y, float *z)
{
/*
  qt_toang: convert quaterion to [angle,axis] representation.
*/
  c_QUAT q;
  float  s, omega;

  qt_normalize (a, &q);
  omega = acos (q.w);
  *ang = omega * 2.0;
  s = sin (omega);
  if (fabs (s) > EPSILON) {
    *x = q.x / s;
    *y = q.y / s;
    *z = q.z / s;
  } else {
    *x = 0.0;
    *y = 0.0;
    *z = 0.0;
  }
}

void qt_make (float w, float x, float y, float z, c_QUAT *out)
{
/*
  qt_make: create quaternion (out = [w,[x,y,z]])
*/
  out->w = w;
  out->x = x;
  out->y = y;
  out->z = z;
}

void qt_identity (c_QUAT *out)
{
/*
  qt_identity: make identity quaternion (out = [1,0,0,0]).
*/
  out->w = 1.0;
  out->x = 0.0;
  out->y = 0.0;
  out->z = 0.0;
}

void qt_zero (c_QUAT *out)
{
/*
  qt_zero: zero quaternion (out = [0,0,0,0]).
*/
  out->w = 0.0;
  out->x = 0.0;
  out->y = 0.0;
  out->z = 0.0;
}

void qt_copy (c_QUAT *a, c_QUAT *out)
{
/*
  qt_copy: quaternion copy (out = a).
*/
  out->w = a->w;
  out->x = a->x;
  out->y = a->y;
  out->z = a->z;
}

void qt_print (c_QUAT *a)
{
/*
  qt_print: print quaternion on stdout.
*/
  printf ("w:%f x:%f y:%f z:%f\n", a->w, a->x, a->y, a->z);
}

void qt_add (c_QUAT *a, c_QUAT *b, c_QUAT *out)
{
/*
  qt_add: quaternion addition (out = a+b).
*/
  c_QUAT temp;

  temp.w = a->w + b->w;
  temp.x = a->x + b->x;
  temp.y = a->y + b->y;
  temp.z = a->z + b->z;
  qt_copy (&temp, out);
}

void qt_sub (c_QUAT *a, c_QUAT *b, c_QUAT *out)
{
/*
  qt_sub: quaterion substraction (out = a-b).
*/
  c_QUAT temp;

  temp.w = a->w - b->w;
  temp.x = a->x - b->x;
  temp.y = a->y - b->y;
  temp.z = a->z - b->z;
  qt_copy (&temp, out);
}

void qt_mul (c_QUAT *a, c_QUAT *b, c_QUAT *out)
{
/*
  qt_multiply: quaternion multiplication (out = a*b).
*/
  c_QUAT temp;

  temp.w = a->w*b->w - a->x*b->x - a->y*b->y - a->z*b->z;
  temp.x = a->w*b->x + a->x*b->w + a->y*b->z - a->z*b->y;
  temp.y = a->w*b->y + a->y*b->w + a->z*b->x - a->x*b->z;
  temp.z = a->w*b->z + a->z*b->w + a->x*b->y - a->y*b->x;
  qt_copy (&temp, out);
}

void qt_scale (c_QUAT *a, float s, c_QUAT *out)
{
/*
  qt_scale: scale quaternion to a new length.
*/
  c_QUAT temp;
  float  len;

  len = (a->x*a->x + a->y*a->y + a->z*a->z + a->w*a->w);
  if (len == 0.0) {
    out->w = 1.0;
    out->x = 0.0;
    out->y = 0.0;
    out->z = 0.0;
    return;
  }
  temp.w = a->w * s / len;
  temp.x = a->x * s / len;
  temp.y = a->y * s / len;
  temp.z = a->y * s / len;
  qt_copy (&temp, out);
}

void qt_div (c_QUAT *a, c_QUAT *b, c_QUAT *out)
{
/*
  qt_div: quaterion division (out = a/b).
*/
  c_QUAT q, t, s, temp;
  float  p;

  qt_copy (b, &q);
  q.x = -q.x;
  q.y = -q.y;
  q.z = -q.z;
  qt_mul (a, &q, &t);
  qt_mul (&q, &q, &s);
  temp.w = t.w / s.w;
  temp.x = t.x / s.w;
  temp.y = t.y / s.w;
  temp.z = t.z / s.w;
  qt_copy (&temp, out);
}

void qt_square (c_QUAT *a, c_QUAT *out)
{
/*
  qt_square: quaternion square (out = a^2).
*/
  float  s;
  c_QUAT temp;

  s = 2 * a->w;
  temp.w = (a->w*a->w - a->x*a->x - a->y*a->y - a->z*a->z);
  temp.x = s * a->x;
  temp.y = s * a->y;
  temp.z = s * a->z;
  qt_copy (&temp, out);
}

void qt_sqrt (c_QUAT *a, c_QUAT *out)
{
/*
  qt_sqrt: quaternion square root (out = sqrt(a)).
*/
  float  len, m;
  float  A, B;
  c_QUAT r, temp;
  
  len = sqrt (a->w*a->w + a->x*a->x + a->y*a->y);
  r.w = a->w / len;
  r.x = a->x / len;
  r.y = a->y / len;
  r.z = 0.0;
  m = 1 / sqrt (r.w*r.w + r.x*r.x);
  A = sqrt ((1 + r.y) / 2);
  B = sqrt ((1 - r.y) / 2);
  temp.w = sqrt (len) * B * r.w * m;
  temp.x = sqrt (len) * B * r.x * m;
  temp.y = sqrt (len) * A;
  temp.z = a->z;
  qt_copy (&temp, out);
}

void qt_normalize (c_QUAT *a, c_QUAT *out)
{
/*
  qt_normalize: quaternion normalization.
*/
  float  len;
  c_QUAT temp;

  len = sqrt (a->x*a->x + a->y*a->y + a->z*a->z + a->w*a->w);
  if (len == 0.0) {
    out->w = 1.0;
    out->x = 0.0;
    out->y = 0.0;
    out->z = 0.0;
    return;
  }
  temp.w = a->w / len;
  temp.x = a->x / len;
  temp.y = a->y / len;
  temp.z = a->z / len;
  qt_copy (&temp, out);
}

void qt_inverse (c_QUAT *a, c_QUAT *out)
{
/*
  qt_inverse: form multiplicative inverse of quaternion.
*/
  float  len;
  c_QUAT temp;

  len = sqrt (a->x*a->x + a->y*a->y + a->z*a->z + a->w*a->w);
  if (len == 0.0) {
    out->w = 1.0;
    out->x = 0.0;
    out->y = 0.0;
    out->z = 0.0;
    return;
  }
  temp.w =  a->w / len;
  temp.x = -a->x / len;
  temp.y = -a->y / len;
  temp.z = -a->z / len;
  qt_copy (&temp, out);
}

void qt_slerp (c_QUAT *a, c_QUAT *b, float alpha, c_QUAT *out)
{
/*
  qt_slerp: spherical linear interpolation of quaternions.
*/
  float  beta, theta;
  float  sin_t, cos_t;
  float  phi;
  int    bflip;
  c_QUAT temp;

  cos_t = a->x*b->x + a->y*b->y + a->z*b->z + a->w*b->w;
  if (cos_t < 0.0) {
    cos_t = -cos_t;
    bflip = 1;
  } else bflip = 0;
  if (1.0 - cos_t < EPSILON) beta = 1.0 - alpha; else {
    theta = acos (cos_t);
    phi = theta;
    sin_t = sin (theta);
    beta = sin (theta - alpha * phi) / sin_t;
    alpha = sin (alpha * phi) / sin_t;
  }
  if (bflip) alpha = -alpha;
  temp.x = beta*a->x + alpha*b->x;
  temp.y = beta*a->y + alpha*b->y;
  temp.z = beta*a->z + alpha*b->z;
  temp.w = beta*a->w + alpha*b->w;
  qt_copy (&temp, out);
}

void qt_slerpspin (c_QUAT *a, c_QUAT *b, float alpha, int spin, c_QUAT *out)
{
/*
  qt_slerpspin: spherical linear interpolation of quaternions
                with extra spins.
*/
  float  beta, theta;
  float  sin_t, cos_t;
  float  phi;
  int    bflip;
  c_QUAT temp;

  cos_t = a->x*b->x + a->y*b->y + a->z*b->z + a->w*b->w;
  if (cos_t < 0.0) {
    cos_t = -cos_t;
    bflip = 1;
  } else bflip = 0;
  if (1.0 - cos_t < EPSILON) beta = 1.0 - alpha; else {
    theta = acos (cos_t);
    phi = theta + spin * M_PI;
    sin_t = sin (theta);
    beta = sin (theta - alpha * phi) / sin_t;
    alpha = sin (alpha * phi) / sin_t;
  }
  if (bflip) alpha = -alpha;
  temp.x = beta*a->x + alpha*b->x;
  temp.y = beta*a->y + alpha*b->y;
  temp.z = beta*a->z + alpha*b->z;
  temp.w = beta*a->w + alpha*b->w;
  qt_copy (&temp, out);
}

void qt_matrix (c_QUAT *a, c_MATRIX mat)
{
/*
  qt_matrix: convert quaternion to rotation matrix.

      ( 1-yy-zz , xy+wz   , xz-wy   )
  T = ( xy-wz   , 1-xx-zz , yz+wx   )
      ( xz+wy   , yz-wx   , 1-xx-yy )
*/
  float x2, y2, z2, wx, wy, wz,
        xx, xy, xz, yy, yz, zz;

  x2 = a->x + a->x; y2 = a->y + a->y; z2 = a->z + a->z;
  wx = a->w * x2;   wy = a->w * y2;   wz = a->w * z2;
  xx = a->x * x2;   xy = a->x * y2;   xz = a->x * z2;
  yy = a->y * y2;   yz = a->y * z2;   zz = a->z * z2;

  mat[X][X] = 1.0 - (yy + zz);
  mat[X][Y] = xy + wz;
  mat[X][Z] = xz - wy;
  mat[X][W] = 0.0;
  mat[Y][X] = xy - wz;
  mat[Y][Y] = 1.0 - (xx + zz);
  mat[Y][Z] = yz + wx;
  mat[Y][W] = 0.0;
  mat[Z][X] = xz + wy;
  mat[Z][Y] = yz - wx;
  mat[Z][Z] = 1.0 - (xx + yy);
  mat[Z][W] = 0.0;
}

void qt_invmatrix (c_QUAT *a, c_MATRIX mat)
{
/*
  qt_invmatrix: convert quaternion to inversed rotation matrix.

      ( 1-yy-zz , xy-wz   , xz+wy   )
  T = ( xy+wz   , 1-xx-zz , yz-wx   )
      ( xz-wy   , yz+wx   , 1-xx-yy )
*/
  float x2, y2, z2, wx, wy, wz,
        xx, xy, xz, yy, yz, zz;

  x2 = a->x + a->x; y2 = a->y + a->y; z2 = a->z + a->z;
  wx = a->w * x2;   wy = a->w * y2;   wz = a->w * z2;
  xx = a->x * x2;   xy = a->x * y2;   xz = a->x * z2;
  yy = a->y * y2;   yz = a->y * z2;   zz = a->z * z2;

  mat[X][X] = 1.0 - (yy + zz);
  mat[X][Y] = xy - wz;
  mat[X][Z] = xz + wy;
  mat[X][W] = 0.0;
  mat[Y][X] = xy + wz;
  mat[Y][Y] = 1.0 - (xx + zz);
  mat[Y][Z] = yz - wx;
  mat[Y][W] = 0.0;
  mat[Z][X] = xz - wy;
  mat[Z][Y] = yz + wx;
  mat[Z][Z] = 1.0 - (xx + yy);
  mat[Z][W] = 0.0;
}

void qt_frommat (c_MATRIX mat, c_QUAT *out)
{
/*
  qt_frommat: convert rotation matrix to quaternion.
*/
  float  tr, s;
  float  q[4];
  int    i, j, k;
  static int nxt[3] = {Y,Z,X};

  tr = mat[X][X] + mat[Y][Y] + mat[Z][Z];
  if (tr > 0.0) {
    s = sqrt (tr + 1.0);
    out->w = s / 2.0;
    s = 0.5 / s;
    out->x = (mat[Y][Z] - mat[Z][Y]) * s;
    out->y = (mat[Z][X] - mat[X][Z]) * s;
    out->z = (mat[X][Y] - mat[Y][X]) * s;
  } else {
    i = X;
    if (mat[Y][Y] > mat[X][X]) i = Y;
    if (mat[Z][Z] > mat[i][i]) i = Z;
    j = nxt[i];
    k = nxt[j];
    s = sqrt ((mat[i][i] - (mat[j][j]+mat[k][k])) + 1.0);
    q[i] = s / 2.0;
    if (s != 0.0) s = 0.5 / s;
    q[W] = (mat[j][k] - mat[k][j]) * s;
    q[j] = (mat[i][j] + mat[j][i]) * s;
    q[k] = (mat[i][k] + mat[k][i]) * s;
    out->w = q[W];
    out->x = q[X];
    out->y = q[Y];
    out->z = q[Z];
  }
}
