/*SPLINE.C********************************************************************

   Clax: Portable keyframing library
         see inclosed LICENSE.TXT for licensing terms.

         for documentation, refer to CLAX.TXT

   author           : Adept
   file created     : 17/04/97
   file description : spline interpolation routines

   revision history :
     ----- (06/03/97) Adept:  Initial version (spline_1).
     v0.10 (17/04/97) Borzom: Changed to fit clax structure.
           (19/05/97) Borzom: Added 2-key spline support.

   notes            :
     It is Kochanek-Bartels spline sub-type of hermit curve developped
     especially for computer key animations, first presented at SIGGRAPH'84.

*****************************************************************************/

#include <math.h>
#include <stdio.h>
#include "clax.h"
#include "claxi.h"

static float Ease (float t, float a, float b)
{
  float k;
  float s = a+b;

  if (s == 0.0) return t;
  if (s > 1.0) {
    a = a/s;
    b = b/s;
  }
  k = 1.0/(2.0-a-b);
  if (t < a) return ((k/a)*t*t);
    else {
      if (t < 1.0-b) return (k*(2*t-a));
        else {
          t = 1.0-t;
          return (1.0-(k/b)*t*t);
        }
    }
}

static void CompElementDeriv (float pp,  float p,   float pn,
                              float *ds, float *dd, float ksm,
                              float ksp, float kdm, float kdp)
{
  float delm, delp;

  delm = p - pp;
  delp = pn - p;
  *ds  = ksm*delm + ksp*delp;
  *dd  = kdm*delm + kdp*delp;
}

static void CompDeriv (t_KEY *keyp, t_KEY *key, t_KEY *keyn)
{
  float tm, cm, cp, bm, bp, tmcm, tmcp;
  float ksm, ksp, kdm, kdp, delm, delp;
  float dt, fp, fn, c;

  dt = 0.5 * (keyn->frame - keyp->frame);
  fp = (key->frame - keyp->frame) / dt;
  fn = (keyn->frame - key->frame) / dt;
  c  = fabs (key->cont);
  fp = fp + c - c * fp;
  fn = fn + c - c * fn;
  cm = 1.0 - key->cont;
  tm = 0.5 * (1.0 - key->tens);
  cp = 2.0 - cm;
  bm = 1.0 - key->bias;
  bp = 2.0 - bm;
  tmcm = tm * cm;
  tmcp = tm * cp;
  ksm = tmcm * bp * fp; ksp = tmcp * bm * fp;
  kdm = tmcp * bp * fn; kdp = tmcm * bm * fn;
  CompElementDeriv (keyp->val._quat.w, key->val._quat.w, keyn->val._quat.w,
                    &key->dsa, &key->dda, ksm, ksp, kdm, kdp);
  CompElementDeriv (keyp->val._quat.x, key->val._quat.x, keyn->val._quat.x,
                    &key->dsb, &key->ddb, ksm, ksp, kdm, kdp);
  CompElementDeriv (keyp->val._quat.y, key->val._quat.y, keyn->val._quat.y,
                    &key->dsc, &key->ddc, ksm, ksp, kdm, kdp);
  CompElementDeriv (keyp->val._quat.z, key->val._quat.z, keyn->val._quat.z,
                    &key->dsd, &key->ddd, ksm, ksp, kdm, kdp);
}

static void CompDerivFirst (t_KEY *key, t_KEY *keyn, t_KEY *keynn)
{
  float f20, f10, v20, v10;

  f20 = keynn->frame - key->frame;
  f10 = keyn->frame - key->frame;
  v20 = keynn->val._quat.w - key->val._quat.w;
  v10 = keyn->val._quat.w - key->val._quat.w;
  key->dda = (1-key->tens)*(v20*(0.25-f10/(2*f20))+(v10-v20/2)*3/2+v20/2);
  v20 = keynn->val._quat.x - key->val._quat.x;
  v10 = keyn->val._quat.x - key->val._quat.x;
  key->ddb = (1-key->tens)*(v20*(0.25-f10/(2*f20))+(v10-v20/2)*3/2+v20/2);
  v20 = keynn->val._quat.y - key->val._quat.y;
  v10 = keyn->val._quat.y - key->val._quat.y;
  key->ddc = (1-key->tens)*(v20*(0.25-f10/(2*f20))+(v10-v20/2)*3/2+v20/2);
  v20 = keynn->val._quat.z - key->val._quat.z;
  v10 = keyn->val._quat.z - key->val._quat.z;
  key->ddd = (1-key->tens)*(v20*(0.25-f10/(2*f20))+(v10-v20/2)*3/2+v20/2);
}

static void CompDerivLast (t_KEY *keypp, t_KEY *keyp, t_KEY *key)
{
  float f20, f10, v20, v10;

  f20 = key->frame - keypp->frame;
  f10 = key->frame - keyp->frame;
  v20 = key->val._quat.w - keypp->val._quat.w;
  v10 = key->val._quat.w - keyp->val._quat.w;
  key->dsa = (1-key->tens)*(v20*(0.25-f10/(2*f20))+(v10-v20/2)*3/2+v20/2);
  v20 = key->val._quat.x - keypp->val._quat.x;
  v10 = key->val._quat.x - keyp->val._quat.x;
  key->dsb = (1-key->tens)*(v20*(0.25-f10/(2*f20))+(v10-v20/2)*3/2+v20/2);
  v20 = key->val._quat.y - keypp->val._quat.y;
  v10 = key->val._quat.y - keyp->val._quat.y;
  key->dsc = (1-key->tens)*(v20*(0.25-f10/(2*f20))+(v10-v20/2)*3/2+v20/2);
  v20 = key->val._quat.z - keypp->val._quat.z;
  v10 = key->val._quat.z - keyp->val._quat.z;
  key->dsd = (1-key->tens)*(v20*(0.25-f10/(2*f20))+(v10-v20/2)*3/2+v20/2);
}

static void CompDerivTwo (t_KEY *key)
{
  t_KEY *keyn = key->next;
  
  key->dsa = 0.0;
  key->dsb = 0.0;
  key->dsc = 0.0;
  key->dsd = 0.0;
  key->dda = (keyn->val._quat.w - key->val._quat.w) * (1 - key->tens);
  key->ddb = (keyn->val._quat.x - key->val._quat.x) * (1 - key->tens);
  key->ddc = (keyn->val._quat.y - key->val._quat.y) * (1 - key->tens);
  key->ddd = (keyn->val._quat.z - key->val._quat.z) * (1 - key->tens);
  keyn->dda = 0.0;
  keyn->ddb = 0.0;
  keyn->ddc = 0.0;
  keyn->ddd = 0.0;
  keyn->dsa = (keyn->val._quat.w - key->val._quat.w) * (1 - keyn->tens);
  keyn->dsb = (keyn->val._quat.x - key->val._quat.x) * (1 - keyn->tens);
  keyn->dsc = (keyn->val._quat.y - key->val._quat.y) * (1 - keyn->tens);
  keyn->dsd = (keyn->val._quat.z - key->val._quat.z) * (1 - keyn->tens);
}

int spline_init (t_KEY *keys)
{
/*
  spline_init: initialize spline key list.
*/
  int   count;
  t_KEY *curr;

  if (!keys) return clax_err_nullptr;
  if (!keys->next) return clax_err_spline;

  if (keys->next->next) { /* 3 or more keys */
    for (curr = keys->next; curr->next; curr = curr->next)
      CompDeriv (curr->prev, curr, curr->next);
    CompDerivFirst (keys, keys->next, keys->next->next);
    CompDerivLast (curr->prev->prev, curr->prev, curr);
  } else CompDerivTwo (keys); /* 2 keys */
  return clax_err_ok;
}

int spline_getkey_float (t_KEY *keys, float frame, float *out)
{
/*
  spline_getkey_float: get spline interpolated float for frame "frame".
*/
  float t, t2, t3;
  float h[4];

  if (frame < 0.0) return clax_err_badframe;
  if (!keys) return clax_err_nullptr;

  while (frame > keys->next->frame && keys->next) keys = keys->next;
  if (!keys->next || frame < keys->frame) { /* frame is above last key */
    *out = keys->val._float;
    return clax_err_ok;
  }
  t = (frame - keys->frame) / (keys->next->frame - keys->frame);
  t = Ease (t, keys->easefrom, keys->next->easeto);
  t2 = t * t;
  t3 = t2 * t;
  h[0] =  2 * t3 - 3 * t2 + 1;
  h[1] = -2 * t3 + 3 * t2;
  h[2] = t3 - 2 * t2 + t;
  h[3] = t3 - t2;
  *out = (h[0]*keys->val._float) + (h[1]*keys->next->val._float) +
         (h[2]*keys->dda) +        (h[3]*keys->next->dsa);
  return clax_err_ok;
}

int spline_getkey_vect (t_KEY *keys, float frame, c_VECTOR *out)
{
/*
  spline_getkey_vect: get spline interpolated vector for frame "frame".
*/
  float t, t2, t3;
  float h[4];

  if (frame < 0.0) return clax_err_badframe;
  if (!keys) return clax_err_nullptr;

  while (frame > keys->next->frame && keys->next) keys = keys->next;
  if (!keys->next || frame < keys->frame) { /* frame is above last key */
    vec_copy (&keys->val._vect, out);
    return clax_err_ok;
  }
  t = (frame - keys->frame) / (keys->next->frame - keys->frame);
  t = Ease (t, keys->easefrom, keys->next->easeto);
  t2 = t * t;
  t3 = t2 * t;
  h[0] =  2 * t3 - 3 * t2 + 1;
  h[1] = -2 * t3 + 3 * t2;
  h[2] = t3 - 2 * t2 + t;
  h[3] = t3 - t2;
  out->x = (h[0]*keys->val._vect.x) + (h[1]*keys->next->val._vect.x) +
           (h[2]*keys->dda) +         (h[3]*keys->next->dsa);
  out->y = (h[0]*keys->val._vect.y) + (h[1]*keys->next->val._vect.y) +
           (h[2]*keys->ddb) +         (h[3]*keys->next->dsb);
  out->z = (h[0]*keys->val._vect.z) + (h[1]*keys->next->val._vect.z) +
           (h[2]*keys->ddc) +         (h[3]*keys->next->dsc);
  return clax_err_ok;
}

int spline_getkey_quat (t_KEY *keys, float frame, c_QUAT *out)
{
/*
  spline_getkey_quat: get spline interpolated quaternion for frame "frame".
*/
  float t, t2, t3;
  float h[4];

  if (frame < 0.0) return clax_err_badframe;
  if (!keys) return clax_err_nullptr;

  while (frame > keys->next->frame && keys->next) keys = keys->next;
  if (!keys->next || frame < keys->frame) { /* frame is above last key */
    qt_copy (&keys->val._quat, out);
    return clax_err_ok;
  }
  t = (frame - keys->frame) / (keys->next->frame - keys->frame);
  t = Ease (t, keys->easefrom, keys->next->easeto);
  t2 = t * t;
  t3 = t2 * t;
  h[0] =  2 * t3 - 3 * t2 + 1;
  h[1] = -2 * t3 + 3 * t2;
  h[2] = t3 - 2 * t2 + t;
  h[3] = t3 - t2;
  out->w = (h[0]*keys->val._quat.w) + (h[1]*keys->next->val._quat.w) +
           (h[2]*keys->dda) +         (h[3]*keys->next->dsa);
  out->x = (h[0]*keys->val._quat.x) + (h[1]*keys->next->val._quat.x) +
           (h[2]*keys->ddb) +         (h[3]*keys->next->dsb);
  out->y = (h[0]*keys->val._quat.y) + (h[1]*keys->next->val._quat.y) +
           (h[2]*keys->ddc) +         (h[3]*keys->next->dsc);
  out->z = (h[0]*keys->val._quat.z) + (h[1]*keys->next->val._quat.z) +
           (h[2]*keys->ddd) +         (h[3]*keys->next->dsd);
  return clax_err_ok;
}
