// mighty.engine version 0.10
// header file

#ifndef _ENGINE_H
#define _ENGINE_H

#include <stdio.h>
#include <math.h>

#define PI             3.141592687
#define sf_dot         1            // surface
#define sf_wireframe   2
#define sf_plain       3
#define sf_texture     4
#define sf_mirror      5
#define sh_ambient     1            // shading
#define sh_lambert     2
#define sh_gouraud     3
#define sh_environment 4
#define sh_phong       5
#define sh_bump        6
#define pf_transparent 1            // polygon flags
#define pf_perspective 2
#define pf_antialias   4
#define pf_twosided    8
#define af_shadows     1            // atmosphere flags
#define af_fog         2
#define af_haze        4
#define s_zmin         1
#define s_zmid         2
#define s_zmax         3
#define MAXLIGHT       32
#define MAXKEYS        512
#define CLIPDISTANCE   2.0

#define ABS(a)          ((a < 0) ? -a : a)
#define SGN(a)          ((a < 0) ? -1 : 1)
#define MAX(a,b)        ((a > b) ? a : b)
#define MIN(a,b)        ((a < b) ? a : b)
#define SQRT(a)         ((a > 0.0) ? sqrt(a) : 1)
#define DotProduct(a,b) ((a->x3d*b->x3d)+(a->y3d*b->y3d)+(a->z3d*b->z3d))

enum {                              // 3DS fileformat chunks
  CHUNK_RGBF          = 0x0010,
  CHUNK_RGBB          = 0x0011,
  CHUNK_PRJ           = 0xC23D,
  CHUNK_MLI           = 0x3DAA,
  CHUNK_MAIN          = 0x4D4D,
  CHUNK_OBJMESH       = 0x3D3D,
  CHUNK_BKGCOLOR      = 0x1200,
  CHUNK_AMBCOLOR      = 0x2100,
  CHUNK_OBJBLOCK      = 0x4000,
  CHUNK_TRIMESH       = 0x4100,
  CHUNK_VERTLIST      = 0x4110,
  CHUNK_VERTFLAGS     = 0x4111,
  CHUNK_FACELIST      = 0x4120,
  CHUNK_FACEMAT       = 0x4130,
  CHUNK_MAPLIST       = 0x4140,
  CHUNK_SMOOLIST      = 0x4150,
  CHUNK_TRMATRIX      = 0x4160,
  CHUNK_MESHCOLOR     = 0x4165,
  CHUNK_TXTINFO       = 0x4170,
  CHUNK_LIGHT         = 0x4600,
  CHUNK_SPOTLIGHT     = 0x4610,
  CHUNK_CAMERA        = 0x4700,
  CHUNK_HIERARCHY     = 0x4F00,
  CHUNK_VIEWPORT      = 0x7001,
  CHUNK_MATERIAL      = 0xAFFF,
  CHUNK_MATNAME       = 0xA000,
  CHUNK_AMBIENT       = 0xA010,
  CHUNK_DIFFUSE       = 0xA020,
  CHUNK_SPECULAR      = 0xA030,
  CHUNK_TEXTURE       = 0xA200,
  CHUNK_BUMPMAP       = 0xA230,
  CHUNK_MAPFILE       = 0xA300,
  CHUNK_KEYFRAMER     = 0xB000,
  CHUNK_AMBIENTKEY    = 0xB001,
  CHUNK_TRACKINFO     = 0xB002,
  CHUNK_TRACKOBJNAME  = 0xB010,
  CHUNK_TRACKPIVOT    = 0xB013,
  CHUNK_TRACKPOS      = 0xB020,
  CHUNK_TRACKROTATE   = 0xB021,
  CHUNK_TRACKSCALE    = 0xB022,
  CHUNK_TRACKMORPH    = 0xB026,
  CHUNK_TRACKHIDE     = 0xB029,
  CHUNK_OBJNUMBER     = 0xB030,
  CHUNK_TRACKCAMERA   = 0xB003,
  CHUNK_TRACKFOV      = 0xB023,
  CHUNK_TRACKROLL     = 0xB024,
  CHUNK_TRACKCAMTGT   = 0xB004,
  CHUNK_TRACKLIGHT    = 0xB005,
  CHUNK_TRACKLIGTGT   = 0xB006,
  CHUNK_TRACKSPOTL    = 0xB007,
  CHUNK_FRAMES        = 0xB008,
};

typedef unsigned char  byte;
typedef unsigned short word;
typedef unsigned long  dword;
typedef signed char    sbyte;
typedef signed short   sword;
typedef signed long    sdword;

typedef struct {
  word    id;
  dword   len;
} TChunkHeader, *PChunkHeader;

typedef struct {
  float w, x, y, z;
} quat;

typedef struct {
  float x, y, z;
} Vector;

typedef struct {
  float  frame;
  quat   pos;
  quat   ds, dd;
  float  tens;
  float  cont;
  float  bias;
  float  easeto;
  float  easefrom;
} Key;

typedef struct
{
  int x,  y;
  int z,  c;
  int u,  v;
  int u1, v1;
} Vertex;

typedef struct {
  float x, y, z;
  int   x2, y2;
  int   gc, c;
  int   u1;
  int   v1;
} ClipPoint;

typedef struct {
  Vector *source;
  Vector *dest;
  float  *roll;
} Fcamera;

typedef struct {
  quat   *rotate;
  Vector *scale;
  Vector *position;
  Vector pivot;
} Fobject;

typedef struct {
  Vector *position;
} Flight;

class Point {
public:
  float x3d, y3d, z3d;
  float eyeX, eyeY, eyeZ;
  float localX, localY, localZ;
  float normalX, normalY, normalZ;
  int   x2d, y2d, numNormals;
  long  u, v;
  Point *normal;

  Point ();
 ~Point ();
  Point (float x, float y, float z);

  void Save ();
  void SetCoords (float x, float y, float z);
  void SetTexCoords (float U, float V);
  void Normalise ();
  void ZeroNormal ();
  void AddNormal (float x, float y, float z);
  void AverageNormal ();
};

class Polygon {
public:
  Point      *p1, *p2, *p3, *normal;
  char       *texture, *shadetex;
  char       color;
  int        surface, shading, flags;
  int        midZ;

  Polygon ();
 ~Polygon ();
  Polygon (Point *v1, Point *v2, Point *v3);

  void  SetTo (Point *P1, Point *P2, Point *P3);
  void  CalcNormal ();
  void  SetShading (int surf, int shade, int flag, int c, char *tex, char *tex2);
  void  SetNewLocalOrigin (Point *p);
  void  CalcGNormals ();
  void  Render (ClipPoint *a, ClipPoint *b, ClipPoint *c);
  void  ZRender (ClipPoint *a, ClipPoint *b, ClipPoint *c);
  void  Display ();
};

class Object {
public:
  Point   **point;
  Point   **normal;
  Polygon **poly;
  int     numPoints, numNormals, numPolys, count;
  float   trmat[9];
  Vector  trans;
  Vector  pivot;
  Vector  scal;

  Object ();
 ~Object ();

  void Save ();
  void SetShading (int surf, int shade, int flag, int c, char *tex, char *tex2);
  void SetFlipMatrix (float mat[9]);
  void AddPoint (float x, float y, float z);
  void AddPoint (Point *p);
  void AddNormal (float x, float y, float z);
  void AddNormal (Point *p);
  void AddPoly (Polygon *pg);
  void AddPoly (int p1, int p2, int p3);
  void SetMatrix (float *mat);
  void SetPivot (Vector piv);
  void SetScale (Vector s);
  void SetPosition (Vector pos);
  void SetPosition (float x, float y, float z);
  void TransformLocal ();
  void CalcVertNormals ();
  void AddMidPoint ();
  void EyeSpace ();
};

class Camera {
public:
  Point *target, *position;
  float Nx, Ny, Nz,
        Ux, Uy, Uz,
        Vx, Vy, Vz;

  Camera ();
 ~Camera ();
  Camera (float ViewX, float ViewY, float ViewZ,
	  float TargX, float TargY, float TargZ);

  void SetTo (float ViewX, float ViewY, float ViewZ,
	      float TargX, float TargY, float TargZ);
  void Update ();
};

class Keyframer {
public:
  Fobject **object;
  Flight  **light;
  Fcamera *camera;
  int     numFrames, Frame;

  Keyframer (int frames);
 ~Keyframer ();

  void Prepare ();
  void NextFrame ();
  void SetFrame (int Frame);
};

class Scene {
public:
  Object    **object;
  Point     **light;
  Camera    **cameras;
  Polygon   **sortbuf;
  Camera    *camera;
  Keyframer *keyframer;
  float     AmbientLight, MaxIntensity;
  int       numObjects;
  int       numLights;
  int       numCameras;
  int       numSort;
  int       flags, sorttype;
  int       doscale;

  Scene ();
 ~Scene ();

  void SetActive ();
  void AddObject ();
  void AddLight (float x, float y, float z);
  void AddCamera (float vx, float vy, float vz,
		  float nx, float ny, float nz);
  void EyeSpace ();
  void Display ();
};

extern int       XPROJ;
extern int       YPROJ;
extern int       CLIPX;
extern int       CLIPY;
extern int       MAXX;
extern int       MAXY;
extern int       asine_table [256];
extern char      *ActiveTexture;
extern char      *VirtualScreen;
extern char      *LightTable;
extern Point     *tPoint;
extern Scene     *scene;
extern Polygon   **sortbuffer;

void  InitEngine (int xRes, int yRes);
void  DoneEngine ();
void  InitKeys ();
void  AddKey (float frame, float x, float y, float z, float w,
	      float t, float c, float b, float eto, float efrom);
void  DrawSpline3d (Vector *frame);
void  DrawSpline4d (quat *frame);
int   LoadScene (char *filename);
void  SubVec (Point *p1, Point *p2);
void  qt_matrix (quat q, float *mat);
void  qt_invmatrix (quat q, float *mat);
void  qt_convert (float w, float x, float y, float z, quat *q);
void  qt_mul (quat qL, quat qR, quat *qq);
int   GouraudLight (Point *a, Point *b);
void  qSort (int l, int r);
void  quickSort (int size);
int   ZClip (ClipPoint *src, ClipPoint *dst, int code);

//    graphics primitives
void  DrawLine (int x1, int y1, int x2, int y2, int c);
void  polyDot (ClipPoint *a, ClipPoint *b, ClipPoint *c);
void  polyWire (ClipPoint *a, ClipPoint *b, ClipPoint *c);
void  polyFlat (ClipPoint *a, ClipPoint *b, ClipPoint *c);
void  polyGouraud (ClipPoint *a, ClipPoint *b, ClipPoint *c);
void  polyTexture (ClipPoint *a, ClipPoint *b, ClipPoint *c, char *bitmap);

#endif
